//A basic but very custom made program to draw waterfall onto framebuffer from fft'd samples received from rtlsdr device
//used on GB3KM ATV repeater raspberry Pi as a simple reverse beacon display as part of the carousel
//created mid 2014 by Rob M0DTS - rob at m0dts.co.uk 

//uses librtlsdr and fftw3 libraries which will need to be installed beforehand
//works on raspberry pi model B (80% cpu) or above.

//to compile:
// gcc rtlfb.c -o rtlfb -lrtlsdr -lfftw3 -lm -O3

//to run:
//sudo ./rtlfb 1000000 0 1296500  1296200 1296800 48 300 0 23cm.bmp


#include <complex.h>
#include <fftw3.h>
#include <rtl-sdr.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h> //memset
#include <sys/mman.h>
#include <fcntl.h>

#define WIDTH 720	//d1
#define HEIGHT 576
#define Bpp 32/8
#define FRAMESIZE WIDTH*HEIGHT*Bpp		// screen size x bpp/8
#define debug 1

char * framebuffer;
void draw_pixel(int x, int y,int level);
uint8_t *palette;
char *map;



// sets a single pixel to colour
void draw_pixel(int x, int y,int level) {
	int pos = (((WIDTH*y)+x)*Bpp);
	memset(map+pos, palette[level*3+2], 1);
	memset(map+pos+1, palette[level*3+1], 1);
	memset(map+pos+2, palette[level*3], 1);
	memset(map+pos+3, 1, 1);
}



int main(int argc,char *argv[]){	//SR/Centre/SSB/Beacon/Gain

	int retval;
	int devices;
	static rtlsdr_dev_t *dev;
	int samples_read;
	int n;
	uint8_t *buf;
	uint8_t *beacon_buf;
	uint8_t *ssb_buf;
	uint8_t *average;
	int i,a,z;
	int mr,mi;
	float m,angle1;

	
	//variables passed as command line arguments
	int sample_rate = strtol(argv[1],NULL,10);			//1024000;
    int freqoffset = strtol(argv[2],NULL,10);			//1296550;
	int centre_freq = strtol(argv[3],NULL,10);			//1296550;
	int ssb_centre_freq = strtol(argv[4],NULL,10);		//1296200;
	int beacon_centre_freq =strtol(argv[5],NULL,10);	//1296900;
	int gain =strtol(argv[6],NULL,10)*10;				//480
    int fftnum =strtol(argv[7],NULL,10);				//300
    int lo =strtol(argv[8],NULL,10);					//0
    char *background =argv[9];							//23cm.bmp
	
	int mult=8;
	int NUM_AVG = 1 ;	//take n fft's and average, slow!!
	int NUM_POINTS =2048;
	int NUM_SAMPLES = NUM_POINTS*2;
	int wf_width = 512;
	int wf_height = 150;
	float window[NUM_SAMPLES*mult];
	int left_offset = (WIDTH/2)-(wf_width/2);
	int wf1_top_offset = 125;
	int wf2_top_offset = 350;
 	int wf_span = ((sample_rate*1000/NUM_POINTS)*wf_width)/1000;
	float bin_width = (float)sample_rate/(float)NUM_POINTS;
	int ssb_start_bin = (NUM_POINTS/2) -(((centre_freq-ssb_centre_freq)+(wf_span/2))/(float)bin_width);
	int beacon_start_bin = (NUM_POINTS/2) +(((beacon_centre_freq-centre_freq)-(wf_span/2))/(float)bin_width);
	
	
	buf = malloc(NUM_SAMPLES *mult * sizeof(uint8_t));				//buffer for reading in samples from rtlsdr
	beacon_buf = malloc(NUM_POINTS * sizeof(uint8_t)*wf_height);	//buffer to hold beacon waterfall data
	ssb_buf = malloc(NUM_POINTS * sizeof(uint8_t)*wf_height);		//buffer to hold ssb waterfall data
	average = malloc(NUM_POINTS * sizeof(uint8_t)*NUM_AVG);			//buffer to hold fft averaging data
	
	//init fftw3 plan & buffers
	fftw_plan my_plan;
	fftw_complex *in, *out;
	


	if(debug==1){
		printf("rtlfb Configuration:\n");
		printf("Sample Rate: %d\n",sample_rate);
		printf("Freq Offset: %d\n",freqoffset);	
		printf("RTL Centre: %d\n",centre_freq);
		printf("SSB Centre: %d\n",ssb_centre_freq);
		printf("Beacon Centre: %d\n",beacon_centre_freq);
		printf("Gain: %d\n",gain);
		printf("FFT lines to do: %d\n",fftnum);
		printf("LO freq: %d\n",lo);
		printf("Background image: %s\n",background);
		printf("Number of FFT Points:%d\n",NUM_POINTS);
		printf("Waterfall Span:%d\n",wf_span);
		printf("Bin width Hz:%f\n",bin_width);
		printf("SSB Start Bin:%d\n",ssb_start_bin);
		printf("Beacon Start Bin:%d\n\n",beacon_start_bin);
	}

	//setup frame buffer IO
	int frame = open("/dev/fb0", O_RDWR);
	FILE * f;
	palette = malloc(sizeof(char)*256*3);											//allocate memory for palette data
	framebuffer = malloc(sizeof(char)*(FRAMESIZE));									//allocate memory which is same size as framebuffer
	map = (char*)mmap(0,FRAMESIZE, PROT_READ | PROT_WRITE, MAP_SHARED, frame, 0);	//map framebuffer to memory, seems to be quicker..
	
	//read palette file to memory, this gives the coloured waterfall. must be 256x1 size 24bit RGB format
	if (!(f = fopen("palette.bmp","r"))) {
        printf("Error loading palette file...\n");
		exit(0);
	}
	//read in the file, check it is bitmap
	i = fgetc(f);
	if(i == 66){		//check for "B" at start of file
		fseek(f,10,SEEK_SET);
		i = fgetc(f);	//get image start point
		fseek(f,i,SEEK_SET);
		fread(palette, sizeof(char), 256*3, f);	//read image data to palette buffer
	}
	fclose(f);


	//read background file to memory, this is the pre-created bitmap with text and frequency markers which was created in python script	
	//must be same size as framebuffer and 32bit ARGB (BGRA) format
	if(!(f = fopen(background,"r"))){
	    printf("Error loading background file...\n");
		exit(0);
	}
	
	i = fgetc(f);
	//read in the file, check it is bitmap
	if(i == 66){		//check for "B" at start of file
		fseek(f,10,SEEK_SET);
		i = fgetc(f);	//get image start point
		fseek(f,i,SEEK_SET);
		//read image to buffer last line first - bottom left reference for BMP
		//swap BGRA to RGBA for framebuffer format too
		int h,w;
		for(h=HEIGHT;h>0;h--){
			for(w=WIDTH*4;w>0;w-=4){
				fread(framebuffer+((h*(WIDTH*4))-w)+2, sizeof(char), 1, f);//B
				fread(framebuffer+((h*(WIDTH*4))-w)+1, sizeof(char), 1, f);//G
				fread(framebuffer+((h*(WIDTH*4))-w),sizeof(char), 1, f);//R
				framebuffer[((h*(WIDTH*4))-w)+3] = 0;//A
			}
		}
	}
	fclose(f);
	

	memcpy(map,framebuffer,FRAMESIZE);		//write the background data to the memory mapped framebuffer

	//rtlsdr init
	devices = rtlsdr_get_device_count();	//look for rtlsdr devices to use

	//list devices
	for(n=0;n<devices;n++){
		//printf("Device %d: %s\n\n",devices,rtlsdr_get_device_name(n));
	}

	//open the last devce listed
	if (devices>0){
		retval = rtlsdr_open(&dev, n-1);	//open last device, first if only one connected!
	}else{
		exit(0);
		printf("No RTL Devices found...!");
	}

	

	//pass settings to rtlsdr from command line arguments
	retval = rtlsdr_set_sample_rate(dev, sample_rate*1000);
	retval = rtlsdr_set_center_freq(dev, ((centre_freq-lo)*1000)+freqoffset);
	retval = rtlsdr_set_tuner_gain_mode(dev, 1);
	retval = rtlsdr_set_tuner_gain(dev, gain);
	retval = rtlsdr_reset_buffer(dev);		//req!

	//setup fftw3 plan and buffers
	in = (fftw_complex*) fftw_malloc((sizeof(fftw_complex)*(NUM_SAMPLES*mult)/2));
	out = (fftw_complex*) fftw_malloc((sizeof(fftw_complex)*(NUM_SAMPLES*mult)/2));
	my_plan = fftw_plan_dft_1d(NUM_POINTS*mult, in, out, FFTW_FORWARD, FFTW_ESTIMATE);
	






	//blackman harris windowing values
	float a0 = 0.35875;
	float a1 = 0.48829;
	float a2 = 0.14128;
	float a3 = 0.01168;



	int loop=0;				//counter for number of fft's to do
	int inptr=wf_height;	//waterfall moves upwards
	int outptr=0;



	//create blackman harris window
	for(i = 0; i < (NUM_SAMPLES*mult)/2; i++)
	{
		window[i] = a0  -  (a1 * cos( (2.0 * M_PI * i) / (((NUM_SAMPLES*mult)/2) - 1) ))  +  (a2 * cos( (4.0 * M_PI * i) / (((NUM_SAMPLES*mult)/2) - 1) ))    -    (a3 * cos( (6.0 * M_PI * i) / (((NUM_SAMPLES*mult)/2) - 1) ));
		//window[i] = 0.5 * (1 - cos((2*M_PI)*i/(NUM_SAMPLES/2))); //hamming window
	}

	while(loop < fftnum){
		for(z=0;z<NUM_AVG;z++){
			retval = rtlsdr_read_sync(dev, buf, NUM_SAMPLES*mult, &samples_read);
			if (retval < 0) {
				printf("WARNING: sync read failed.\n");
				exit(0);
			}



			//convert buffer from IQ to complex, seems that rtlsdr outputs IQ data as IQIQIQIQIQIQ so ......
			n=0;
			for (i=0;i<((NUM_SAMPLES*mult)/2);i++){
				__real__  in[i] = buf[n]-127;	//sample is 127 for zero signal,so 127 +/-127 ...
				__imag__ in[i] = buf[n+1]-127;
				//printf("%f + %f\n", __real__  in[i], __imag__ in[i]);
				n=n+2;
			}




			//window input samples before fft
			for(i = 0; i < (NUM_SAMPLES*mult)/2; i++)
			{
				__real__ in[i] *= window[i];
				__imag__ in[i] *= window[i];
			}



			//do the fft
			fftw_execute(my_plan);


			//compute magnitude from complex =  sqrt(real^2 + imaginary^2)
			//FFT displays outputs positive half of spectrum first so have to swap to get correct center zero reference.
			a=0;
			for (i=0;i<NUM_POINTS/2;i++){
				float tmp=0;
				for(n=0;n<mult;n++){
					mr = __real__ out[a] * __real__ out[a];
					mi = __imag__ out[a] * __imag__ out[a];

					m = sqrt(mr+mi);
					if(m>255){m=255;}
					if(m>tmp){tmp=m;}
					a++;
				}
				buf[i+(NUM_POINTS/2)] = tmp;
				//printf("bin %d = %f\n",i,m);
			}


			//second half
			a=(NUM_POINTS*mult)/2;
			for (i=NUM_POINTS/2;i<NUM_POINTS;i++){
				float tmp=0;
				for(n=0;n<mult;n++){
				mr = __real__ out[a] * __real__ out[a];
				mi = __imag__ out[a] * __imag__ out[a];

				m = sqrt(mr+mi);
				
				//crude limiter as colours are 8bit in palette
				if(m>255){m=255;}
				if(m>tmp){tmp=m;}
				a++;
				}
				buf[i-(NUM_POINTS/2)] = tmp;
				//printf("bin %d = %f\n",i,m);
			}



			//Moving average for last NUM_AVG FFT's
			for(a=NUM_AVG*NUM_POINTS;a>0;a--){
				average[a] = average[a-NUM_POINTS];
			}
			for(a=0;a<NUM_POINTS;a++){
				average[a] = buf[a];
			}
		}


			int temp[NUM_POINTS];
			
			for(i=0;i<NUM_POINTS;i++){temp[i]=0;}		//reset average

			for(i=0;i<NUM_AVG;i++){
				n=0;
				for(a=i*NUM_POINTS;a<i*NUM_POINTS+NUM_POINTS;a++){
					temp[n] += average[a];
					n++;
				}
			}

			for(i=0;i<NUM_POINTS;i++){
				temp[i] = temp[i]/NUM_AVG;
				buf[i] = temp[i];
			}




			//move magnitude data into two seperate buffers for ssb and beacon segments
			if(inptr > wf_height){inptr=0;}
			if(outptr> wf_height){outptr=0;}

			for(i=0;i<wf_width;i++){
				if(buf[i+beacon_start_bin] > 255){buf[i+beacon_start_bin] = 255;}
				if(buf[i+ssb_start_bin] > 255){buf[i+ssb_start_bin] = 255;}
				beacon_buf[((inptr)*wf_width)+i] = buf[i+beacon_start_bin];
				ssb_buf[((inptr)*wf_width)+i] = buf[i+ssb_start_bin];
			}
			inptr++;
			
			
			//draw pixel to the framebuffer, or in this case to memory mapped framebuffer
			// write array to framebuffer on even lines, stops interlaced video wobble!
			int loc;
			for(z=0;z<wf_height;z++){
				loc = outptr+z;
				if(loc > wf_height){loc = (outptr+z)-wf_height;}
				for(i=0;i<wf_width;i++){
					if(loop % 2){
						draw_pixel(i+left_offset,z+wf1_top_offset,beacon_buf[(loc*wf_width)+i]);
						draw_pixel(i+left_offset,z+wf2_top_offset,ssb_buf[(loc*wf_width)+i]);
					}
				}
			}
			outptr++;
		loop++;
	}


	//clean up everything?
	fftw_destroy_plan(my_plan);
	free(framebuffer);
	free(palette);
	fftw_free(in);
	fftw_free(out);
	free(buf);
	free(beacon_buf);
	free(ssb_buf);
	free(average);

	close(frame);
	munmap(map,FRAMESIZE);
	rtlsdr_close(0);
}
